#ifdef WIN32
	#include "stdafx.h"
#endif

#include "Log.h"

#include <limits>
#include <cstdarg>
using namespace std;

#ifdef WIN32
	#include "rsMemUtils.h"
	#include "Utils.h"
#endif //#ifdef WIN32

Log g_log;

//// ================== ILog interface ===============================
//#ifdef WIN32
//
//void ILog::mbox (const char* format, ...) {
//#ifndef NOLOG
//	va_list args;
//	va_start (args, format);
//	mbox ("Log", format, args);
//	va_end (args);
//#endif
//}
//
//#endif //ifdef WIN32

// ====================== Log class ================================

#ifdef MULTITRHEADED_LOG
	CMutex g_lockMutex (FALSE);
#endif

#ifdef WIN32
void dprintf (const char* format, ...) { //TODO: add to rsUtils
#ifndef NOLOG
	#ifdef MULTITRHEADED_LOG
		CMutexLock lock (g_lockMutex);
	#endif

	va_list args;
	va_start (args, format);
	char buf [1000];
	vsprintf (buf, format, args);
	va_end (args);

	OutputDebugString (buf);
#endif
};
#endif //#ifdef WIN32

Log::Log() {
	Init (NULL, disable);
}

Log::Log (const char* fname, Log::EnableMode mode, bool append, OpenFileMode ofm, int level) 
{
#ifndef NOLOG
	Init (fname, mode, append, ofm, level);
#endif
}

void Log::Init (const char* fname, EnableMode mode, bool append, OpenFileMode ofm, int level) {
	disabled = (mode == disable);
	whenToOpen = ofm;

	if (fname == NULL)
		disabled = true;

	m_level = level;
	m_currLevel = 0; //always print before level(int i) used
	error = false;
	n = 16;
	specChars = "\r\n\t\a";

	if (disabled)
		return;
	
	fileName = fname;
	fileHandle = NULL;
	m_append = append;
	
	if (! m_append) {
		FILE * f = fopen (fname, "w+");
		if (f == NULL) {
			error = true;
			return;
		}
		fclose (f);
	}

	if (whenToOpen == onlyOnce) {
		if ((fileHandle = fopen (fname, "a+")) == NULL)
			error = true;
	}
}

void Log::printf (const char* format, ...) {
#ifndef NOLOG
	#ifdef MULTITRHEADED_LOG
		CMutexLock lock (g_lockMutex);
	#endif

	if (disabled || !checkLevel()) return;

	FILE * f = getFileHandle();
	if (f == NULL) {
		error = true;
		return;
	}

	va_list args;
	va_start (args, format);
	vfprintf (f, format, args);
	va_end (args);

	releaseFileHandle (f);
#endif
};
		
void Log::myputc (char ch) { //prints special chars as "."
#ifndef NOLOG
	if (disabled || !checkLevel()) return;
	if (strchr (specChars, ch))
		printf (".");
	else
		printf ("%c", ch);
#endif
}

void Log::myfputc (char ch, FILE* f) {
#ifndef NOLOG
	if (strchr (specChars, ch))
		::fputc ('.', f);
	else
		::fputc (ch, f);
#endif
}

void Log::dump (void * void_data, size_t size, int what, const char * txtTitle, const char* hexTitle) {
#ifndef NOLOG
	#ifdef MULTITRHEADED_LOG
		CMutexLock lock (g_lockMutex);
	#endif

	if (disabled || !checkLevel()) return;

	const unsigned char* data = (const unsigned char*) void_data;
	FILE * f = getFileHandle();
	if (f == NULL) {
		error = true;
		return;
	}

	size_t i;
	if (what & text) {
		fprintf (f, "\n%s\n", txtTitle);
		for (i = 0; i < size; i++) {
			if (data[i] != '\r')
				fprintf (f, "%c", data[i]);
		}
		fprintf (f, "\n");
	}

	if (what & hex) {
		fprintf (f, "\n%s\n", hexTitle);
		for (i = 0; i < (size/n)*n; i++) {
			if (data [i] < 0x10) fprintf (f, "0");
			fprintf (f, "%X ", data[i]);

			if ((i+1)%n == 0) {
				fprintf (f, "  ");
				for (size_t j = i - n + 1; j <= i; j++)
					myfputc (data[j], f);
				fprintf (f, "\n");
			}
		}

		size_t left = size - i;
		if (left > 0) {
			for (size_t j = 0; j < left; j++) {
				if (data [i + j] < 0x10) fprintf (f, "0");
				fprintf (f, "%X ", data [i + j]);
			}
			for (size_t j = 0; j < n - left; j++)
				fprintf (f, "   "); //3 spaces
			fprintf (f, "  ");
			for (size_t j = 0; j < left; j++)
				myfputc (data[i + j], f);
		} 
		fprintf (f, "\n");
	}
	releaseFileHandle (f);
#endif
}

#ifdef WIN32

void Log::mbox (const char * title, const char* format, ...) {
#ifndef NOLOG
	#ifdef MULTITRHEADED_LOG
		CMutexLock lock (g_lockMutex);
	#endif

	if (disabled || !checkLevel()) return;

	va_list args;
	va_start (args, format);
	char buf [1000];
	vsprintf (buf, format, args);
	::MessageBox (NULL, buf, title, MB_OK);
	va_end (args);
#endif
}

#endif //ifdef WIN32

void Log::logTime (const char* scopeName, double seconds, const char* executionKind) {
	#ifdef MULTITRHEADED_LOG
		CMutexLock lock (g_lockMutex);
	#endif
	if (strcmp (executionKind, "COUNT") != 0)
		printf ("Scope {%s} %s execution time: %.3lf s\n", scopeName, executionKind, seconds/CLOCKS_PER_SEC);
	else
		printf ("Scope {%s} executions COUNT: %.0lf\n", scopeName, seconds); 
}

void Log::logMemory (const char* scopeName, double size, const char* executionKind) {
	#ifdef MULTITRHEADED_LOG
		CMutexLock lock (g_lockMutex);
	#endif
	if (strcmp (executionKind, "COUNT") != 0)
		printf ("Scope {%s} %s memory usage: %.3lf K\n", scopeName, executionKind, size/1024);
	else
		printf ("Scope {%s} executions COUNT: %.0lf\n", scopeName, size); 
}

Log::~Log() {
	if (whenToOpen == onlyOnce && fileHandle != NULL)
		fclose (fileHandle);
}

#ifdef WIN32
// ====================== rsLog class =================================

rsLog::rsLog (const char* fname, bool clearOld) 
	: Log (fname)
{
#ifndef NOLOG
	if(clearOld)
	{
		DeleteFile(fname);
		//FILE * f = fopen (fname, "w+");
		//if (f == NULL) {
		//	error = true;
		//	return;
		//}
		//mbr_bAddTimeStamp=true;
		//fclose (f);
	}
	mbr_bAddTimeStamp=false;
#endif
}
void rsLog::clearlog ()
{
#ifndef NOLOG
	DeleteFile(fileName.c_str());
#endif
}
void rsLog::setfname(const char* fname)
{
#ifndef NOLOG
	fileName = fname;
#endif
}
void rsLog::printfln (const char* format, ...) {
#ifndef NOLOG
	#ifdef MULTITRHEADED_LOG
		CMutexLock lock (g_lockMutex);
	#endif

	va_list args;
	va_start (args, format);
	
	FILE * f = fopen (fileName.c_str(), "a+");
	if (f == NULL) {
		error = true;
		return;
	}
	if(mbr_bAddTimeStamp)
		AddTimeStamp(f);
	vfprintf (f, format, args);
	va_end (args);
	fprintf(f,"\r\n");
	fclose (f);
#endif
}

void rsLog::SetAddTimeStamp(bool addIt)
{
	mbr_bAddTimeStamp=addIt;
}
void rsLog::AddTimeStamp(FILE * f)
{
	string date=Utils::GetLocalDateAndTimeAsString();
	fprintf (f, "[%s]: ", date.c_str());
}

void rsLog::ClearErrorLog (const char* errorFileName) {
	if(errorFileName==NULL)
	{
		DeleteFile("error.log");
	}else{
		DeleteFile(errorFileName);
	}
}
void rsLog::throwException (const char* errorFileName, const char* format, ...) {
	#ifdef MULTITRHEADED_LOG
		CMutexLock lock (g_lockMutex);
	#endif
	
	string fName;
	if(errorFileName==NULL)
		fName="error.log";
	else
		fName=errorFileName;
	va_list args;
	va_start (args, format);
	
	FILE * f = fopen (fName.c_str(), "a+");
	if (f == NULL) {
		return;
	}
	AddTimeStamp(f);
	vfprintf (f, format, args);
	va_end (args);
	fprintf(f,"\r\n");
	fclose (f);
}


// ====================== TimeLog class ================================

TimeLog::TimeLog (const char* scopeName, ITimeLogWriter * logObject)
{
#ifndef NOLOG
	m_logObject = logObject;
	m_logFunc = NULL;
	m_scopeName = scopeName;
	m_startTime = clock ();
#endif
}

TimeLog::TimeLog (const char* scopeName, TimeLogWriterFunctT logFunc)
{
#ifndef NOLOG
	m_logObject = NULL;
	m_logFunc = logFunc;
	m_scopeName = scopeName;
	m_startTime = clock ();
#endif
}

void TimeLog::doLog (double duration, const char* executionKind) {
#ifndef NOLOG
	if (m_logObject != NULL)
		m_logObject->logTime (m_scopeName.c_str(), duration, executionKind);
	else if (m_logFunc != NULL)
		m_logFunc (m_scopeName.c_str(), duration, executionKind);
	else
		printf ("Scope {%s} %s execution time: %.3lf s\n", m_scopeName.c_str(), executionKind, duration/CLOCKS_PER_SEC);
#endif
}


TimeLog::~TimeLog () 
{
#ifndef NOLOG
	if (m_startTime != 0) {
		clock_t endTime = clock();
		double duration = (double)(endTime - m_startTime) / CLOCKS_PER_SEC;

		doLog (duration);
	}
#endif
}
// ====================== MemoryLog class =================================

MemoryLog::MemoryLog (const char* scopeName, IMemoryLogWriter * logObject)
{
#ifndef NOLOG
	m_logObject = logObject;
	m_logFunc = NULL;
	m_scopeName = scopeName;
	m_startMem = getCurrentValue();
#endif
}

MemoryLog::MemoryLog (const char* scopeName, MemoryLogWriterFunctT logFunc)
{
#ifndef NOLOG
	m_logObject = NULL;
	m_logFunc = logFunc;
	m_scopeName = scopeName;
	m_startMem = getCurrentValue();
#endif
}

double MemoryLog::getCurrentValue () {
	return rs::CurrentMemUsage();
}

void MemoryLog::doLog (double size, const char* executionKind) {
#ifndef NOLOG
	if (m_logObject != NULL)
		m_logObject->logMemory (m_scopeName.c_str(), size, executionKind);
	else if (m_logFunc != NULL)
		m_logFunc (m_scopeName.c_str(), size, executionKind);
	else
		printf ("Scope {%s} %s memory usage: %.3lf K\n", m_scopeName.c_str(), executionKind, (double)size/1024);
#endif
}

MemoryLog::~MemoryLog () 
{
#ifndef NOLOG
	if (m_startMem != 0) {
		double endMem = getCurrentValue();
		double memUsage = endMem - m_startMem;
		doLog (memUsage);
	}
#endif
}

// ====================== LogStatsContainer class ================================
#endif //#ifdef WIN32

