#ifndef Log_H
#define Log_H

// When defined no log is created and printed (works much faster; smaller object code).
//#define NOLOG 
// Note: You'd better define it in projects properties not uncomment it here. 
// Why: e.g. cvs issues...

#include <stdio.h>
#include <string>
#include <ctime>
#include <queue>
#include <limits>
using namespace std;

#ifdef WIN32
	#include <windows.h>
	#include <psapi.h>

	void dprintf (const char* format, ...);
#endif

//#define MULTITRHEADED_LOG
struct ITimeLogWriter {
	void logValue (const char* scopeName, double time, const char* executionKind ="") {
		logTime (scopeName, time, executionKind);
	}

	virtual void logTime (const char* scopeName, double time, const char* executionKind ="") = 0;
};

struct IMemoryLogWriter {
	void logValue (const char* scopeName, double size, const char* executionKind = "") {
		logMemory (scopeName, size, executionKind);
	}

	virtual void logMemory (const char* scopeName, double size, const char* executionKind = "") = 0;
};

typedef void (*TimeLogWriterFunctT) (const char* scopeName, double seconds, const char* executionKind);
typedef void (*MemoryLogWriterFunctT) (const char* scopeName, double size, const char* executionKing);

class ILog  {
public:
	enum DisplayMode {hex = 0x01, text = 0x02};
	enum EnableMode {disable = 0, enable = 1};
	enum OpenFileMode {eachTime, onlyOnce};

	virtual void printf (const char * format, ...) = 0;
	virtual void myputc (char ch) = 0;
	virtual void dump (void *data, size_t size, int what = text | hex, const char *txtTitle = "----------- TXT -----------", const char* hexTitle = "----------- HEX -----------") = 0;
	virtual ILog& level(int i) = 0;
	virtual operator bool () = 0;
	virtual void setEnable (EnableMode mode) = 0;
	//virtual void setOpenFileMode (OpenFileMode ofm) = 0;

#ifdef WIN32
	virtual void mbox (const char* title, const char* format, ...) = 0;
#endif
};

class Log : public ILog, public ITimeLogWriter, public IMemoryLogWriter {
protected:
	string fileName;
	int m_level; ///< required level
	int m_currLevel; ///< current level
	bool error;
	const char * specChars;
	bool disabled, m_append;

	OpenFileMode whenToOpen;
	FILE * fileHandle;
   
	void myfputc (char ch, FILE* f);
	FILE* getFileHandle () {
		if (whenToOpen == onlyOnce)	
			return fileHandle;
		else 
			return fopen (fileName.c_str(), "a+");
	}
	void releaseFileHandle (FILE * f) {
		if (whenToOpen == eachTime)
			fclose (f);
	}


	//level 0 - everything; else print only for currLevel <= level
	bool checkLevel () { return m_level == 0 || m_currLevel <= m_level; }

public:
	int n; //bytes per row

	Log ();
	Log (const char* fname, EnableMode mode = enable, bool append = false, OpenFileMode openFile = eachTime, int level = 0);
	void Init (const char* fname, EnableMode mode = enable, bool append = false, OpenFileMode openFile = eachTime, int level = 0);

	virtual void setEnable (EnableMode mode) {
		disabled = (mode == disable);
	}

	//TODO: implement switching on the fly
/*	virtual void setOpenFileMode (OpenFileMode ofm) {
		whenToOpen = ofm;
	}*/

	virtual void printf (const char * format, ...);
	virtual void myputc (char ch);
	virtual void dump (void *data, size_t size, int what = text | hex, const char *txtTitle = "----------- TXT -----------", const char* hexTitle = "----------- HEX -----------");
	virtual ILog& level(int i) { m_currLevel = i; return *this; }
	virtual int getLevel() { return m_level; }
	virtual bool isDisabled() { return disabled; }
	virtual operator bool () { return !error; }

#ifdef WIN32
	virtual void mbox (const char* title, const char* format, ...);
#endif

	virtual void logTime (const char* scopeName, double seconds, const char* executionKind = "");
	virtual void logMemory (const char* scopeName, double size, const char* executionKind = "");

	virtual ~Log ();
};

#ifdef WIN32
class rsLog : public Log {
protected:
	bool mbr_bAddTimeStamp;
	static void AddTimeStamp(FILE * f);
public:
	rsLog (const char* fname, bool clearOld = true);

	void clearlog ();
	void setfname(const char* fname);
	void printfln (const char * format, ...);

	void SetAddTimeStamp(bool addIt);
	static void throwException (const char* errorFileName, const char* format, ...);
	static void ClearErrorLog (const char* errorFileName=NULL);
};


//synopsis:
//	void f (...) {
//		TimeLog __tl ("f function", &log); //log is global Log object
//		//or
//      TimeLog __tl ("f function"); //prints with std::printf
//		...
//	}	--> here __tl will call log.logTime ("f function", duration),
//		where duration is time in seconds taken by f's execution
//	NOTE: Because of unpredictable order of destructors' execution,
//	destructing time of stack allocated objects is almost surely not 
//	measured correctly.

class TimeLog {
protected:
	ITimeLogWriter * m_logObject;
	TimeLogWriterFunctT m_logFunc;

	string m_scopeName;
	double m_startTime;
public:
	typedef ITimeLogWriter log_writer_class;
	typedef TimeLogWriterFunctT log_writer_funct;

	TimeLog (const char* scopeName, ITimeLogWriter * logObject);
	TimeLog (const char* scopeName, TimeLogWriterFunctT logFunc = NULL);
	void doLog (double duration, const char* executionKind = "");
	static double getCurrentValue () { return (double)clock (); }
	~TimeLog ();
};

class MemoryLog {
protected:
	IMemoryLogWriter * m_logObject;
	MemoryLogWriterFunctT m_logFunc;

	string m_scopeName;
	double m_startMem;
public:
	typedef IMemoryLogWriter log_writer_class;
	typedef MemoryLogWriterFunctT log_writer_funct;

	MemoryLog (const char* scopeName, IMemoryLogWriter * logObject);
	MemoryLog (const char* scopeName, MemoryLogWriterFunctT logFunc = NULL);
	void doLog (double size, const char* executionKind = "");
	static double getCurrentValue ();
	~MemoryLog ();
};

enum LogStatsOptions {lsoNone = 0, lsoCount = 1, lsoSum = 2, lsoMin = 4, lsoMax = 8, lsoAverage = 16, lsoDefault = 128,
							lsoAll = lsoCount|lsoSum|lsoMin|lsoMax|lsoAverage};

//synopsis:
//	void f (...) {
//		static TimeLogStats __tl ("f function", &log);
//		TimeLogStatsMeasurement __tlm (__tl);
//		...
//	} //on destruction of static objects (program terminatin/library unload) 
//	  //average time value will be logged
template <class log_type>
class LogStatsContainer {
public:
	typedef typename log_type::log_writer_class log_writer_class;
	typedef typename log_type::log_writer_funct log_writer_funct;
protected:
	static LogStatsOptions s_globalOptions;
	LogStatsOptions m_options;

	log_writer_class * m_logObject;
	log_writer_funct m_logFunct;
	string m_scopeName;
    
	unsigned long m_measureTimes;
	double m_maxValue, m_minValue, m_sumValue;
	queue<double> m_startValues;
public:
	LogStatsContainer (const char* scopeName, log_writer_class * logObject, int options = lsoDefault) {
		#ifndef NOLOG
			m_scopeName = scopeName;
			m_logObject = logObject;
			m_logFunct = NULL;
			m_options = (LogStatsOptions) ((options == lsoDefault) ? s_globalOptions : options);

			m_sumValue = 0;
			m_minValue = (numeric_limits<double>::max)();
			m_maxValue = (numeric_limits<double>::min)();
		#endif	
	}

	LogStatsContainer (const char* scopeName, log_writer_funct logFunc = NULL, int options = lsoDefault) {
		#ifndef NOLOG
			m_scopeName = scopeName;
			m_logObject = NULL;
			m_logFunct = logFunc;
			m_options = (LogStatsOptions) ((options == lsoDefault) ? s_globalOptions : options);

			m_sumValue = 0;
			m_minValue = (numeric_limits<double>::max)();
			m_maxValue = (numeric_limits<double>::min)();
		#endif	
	}

	static void SetGlobalOptions (int options = lsoAll) {
		s_globalOptions = (LogStatsOptions) options;
	}

	void doLog (double value, const char* executionKind = "") {
		#ifndef NOLOG
			if (m_logObject != NULL)
				m_logObject->logValue (m_scopeName.c_str(), value, executionKind);
			else if (m_logFunct != NULL)
				m_logFunct (m_scopeName.c_str(), value, executionKind);
			else
				printf ("Scope {%s} %s execution time: %.3lf\n", m_scopeName.c_str(), executionKind, value);
		#endif
	}

	virtual void start() {
		#ifndef NOLOG
			#ifdef MULTITRHEADED_LOG
				CMutexLock lock (g_lockMutex);
			#endif
				m_startValues.push (log_type::getCurrentValue());
		#endif
	}

	virtual void stop () {
		#ifndef NOLOG
			#ifdef MULTITRHEADED_LOG
				CMutexLock lock (g_lockMutex);
			#endif
			
			double endValue = log_type::getCurrentValue();
			double startValue = m_startValues.front();
			m_startValues.pop();

			double diff = endValue - startValue; 
			m_sumValue += diff;
			if (diff > m_maxValue)
				m_maxValue = diff;
			if (diff < m_minValue)
				m_minValue = diff;
			m_measureTimes++;
		#endif
	}

	~LogStatsContainer () {
		#ifndef NOLOG
			if (m_measureTimes < 1) {
				doLog (m_measureTimes, "COUNT");
				return;
			}

			if (m_options & lsoCount)
				doLog (m_measureTimes, "COUNT");
			if (m_options & lsoSum)
				doLog (m_sumValue, "SUM");
			if (m_options & lsoMin)
				doLog (m_minValue, "MIN");
			if (m_options & lsoMax)
				doLog (m_maxValue, "MAX");
			if (m_options & lsoAverage)
				doLog (m_sumValue/m_measureTimes, "AVERAGE");
		#endif
	}
};

template <class log_type>
LogStatsOptions LogStatsContainer<log_type>::s_globalOptions = lsoAll;

template <class log_type>
class LogStatsMeasurement {
protected:
	LogStatsContainer<log_type> & m_log;
public:
	LogStatsMeasurement (LogStatsContainer<log_type> & log) : m_log (log) {
#ifndef NOLOG
		m_log.start();
#endif
	}
	~LogStatsMeasurement() {
#ifndef NOLOG
		m_log.stop();
#endif
	}
};

#ifdef NOLOG
	#define LogTime(scopeName, log)
	#define LogTimeStats(scopeName, log)
	#define LogMemory(scopeName, log)
	#define LogMemoryStats(scopeName, log)
	#define LogStats(scopeName, log)
#else
	#define LogTime(scopeName, log) \
		TimeLog _tl ((scopeName), &log); 

	#define LogTimeStats(scopeName, log) \
		static LogStatsContainer<TimeLog> __tl ((scopeName), &log); \
		LogStatsMeasurement<TimeLog> __tls (__tl); 

	#define LogMemory(scopeName, log) \
		MemoryLog _tl ((scopeName), &log); 

	#define LogMemoryStats(scopeName, log) \
		static LogStatsContainer<MemoryLog> __tl ((scopeName), &log); \
		LogStatsMeasurement<MemoryLog> __tls (__tl); 

	#define LogStats(scopeName, log) \
		static LogStatsContainer<TimeLog> __tl ((scopeName), &log); \
		LogStatsMeasurement<TimeLog> __tlm (__tl); \
		static LogStatsContainer<MemoryLog> __ml ((scopeName), &log); \
		LogStatsMeasurement<MemoryLog> __mlm (__ml);
#endif
#endif //#ifdef WIN32

extern Log g_log;

#endif
