// Description: 
// Implements classes for accessing VanyoG Bible modules
//
// Author: Radoslav Atanasov <rsdevelop@gmail.com>, (C) 2005
// Copyright: See COPYING file that comes with this distribution

#include "vanyog_bible.h"
#include "qdir.h"

namespace BgBible {
	namespace VanyoG {
		CBibleStructure::CBibleStructure (const char* filePath, unsigned booksCount) throw (ios_base::failure) {
			ifstream fin (filePath);
			if (! fin) 
				throw ios_base::failure (format("CBibleSructure::CBibleSructure(): error opening file %s", filePath));  
			else {
				m_verses.reserve (booksCount);
				for (unsigned bk = 0; bk < booksCount; bk++) {
					unsigned chapCnt, versesCnt;
					
					fin >> chapCnt;
					m_verses.push_back(vector<unsigned>());
					m_verses[bk].reserve (chapCnt);
					
					for (unsigned chap = 0; chap < chapCnt; chap++) {
						fin >> versesCnt;
						m_verses[bk].push_back (versesCnt);
					}
				}
				
				fin.close();
				
				CalcStartNums ();
			}
		}
		
		void CBibleStructure::CalcStartNums () {
			unsigned startNum = 0;
			m_startVerseNums.reserve (m_verses.size());
			
			for (size_t bookID = 0; bookID < m_verses.size(); bookID++) {
				size_t chapCnt = m_verses[bookID].size();

				m_startVerseNums.push_back (vector<unsigned>());
				m_startVerseNums[bookID].reserve (chapCnt);
				
				m_bookStartVerseNums.push_back (startNum);
				for (unsigned chap = 0; chap < chapCnt; chap++) {
					m_startVerseNums [bookID].push_back (startNum);
					startNum += m_verses [bookID][chap];
				}				
			}
			
			m_totalVerseCount = startNum;
		}
		
		unsigned CBibleStructure::toNum (const CRef& ref) const throw (out_of_range) {
			ValidateRef (ref);
			return m_startVerseNums[ref.bookID()][ref.chapter()] + ref.verse();
		}
		
		CRef CBibleStructure::toRef (unsigned num) const throw (out_of_range) {
			ValidateNum (num);
			
			vector<unsigned>::const_iterator bookIdI = lower_bound (m_bookStartVerseNums.begin(), m_bookStartVerseNums.end(), num);
			unsigned bookID = bookIdI - m_bookStartVerseNums.begin();
			if (bookIdI == m_bookStartVerseNums.end() || num < *bookIdI)
				bookID--;
			
			const vector<unsigned> & chapVerses = m_startVerseNums[bookID];
			vector<unsigned>::const_iterator chapI = lower_bound (chapVerses.begin(), chapVerses.end(), num);
			unsigned chapter = chapI - chapVerses.begin();
			if (chapI == chapVerses.end() || num < *chapI)
				chapter--;
			
			return CRef(bookID, chapter, num - chapVerses[chapter]);
		}
		
		CDiff::CDiff (const char * filePath) throw (ios_base::failure) {
			ifstream fin (filePath);
			if (! fin) 
				throw ios_base::failure (format("CDiff::CDiff(): error opening file %s", filePath));  
			else {
				while (fin) {
					unsigned gBook, lChap, lVerse;
					int dChap, dVerse;
					
					fin >> gBook >> lChap >> lVerse;
					fin >> dChap >> dVerse;
					
					m_localToGlobal.insert (
						make_pair (CRef(gBook - 1, lChap - 1, lVerse - 1), 
									CRefDelta(dChap, dVerse)));
					m_globalToLocal.insert (
						make_pair (CRef(gBook - 1, lChap - 1 + dChap, lVerse - 1 + dVerse), 
									CRefDelta(- dChap, - dVerse)));
				}
				fin.close();
			}
		}
		
		CRef CDiff::convertWithMap (const CRef& ref, const DiffMapT& aMap) {
			DiffMapT::const_iterator I = aMap.begin();
			for (; I != aMap.end(); I++) {
				if (I->first.chaptersBeginning() == ref.chaptersBeginning())
					return ref + I->second;
			}
			return ref;
		}
		
		CTitles::CTitles (const char* filePath) throw (ios_base::failure) {
			ifstream fin (filePath);
			if (! fin) 
				throw ios_base::failure (format("CTitles::CTitles(): error opening file %s", filePath));  
			else {
				unsigned booksCnt; 
				fin >> booksCnt;
				m_globalBookIDs.reserve (booksCnt);	m_fullTitles.reserve (booksCnt);	
				m_shortTitles.reserve (booksCnt); m_abbrTitles.reserve (booksCnt);
				
				for (unsigned bk = 0; bk < booksCnt; bk++) {
					int globalID; fin >> globalID; 
					m_globalBookIDs.push_back (globalID - 1);
					m_globalsToLocalsMap.insert (make_pair(globalID - 1, bk));
				}
				
				char title [MAX_TITLE+1];
				fin.getline (title, MAX_TITLE); //dummy - skip '\n' at the end of the first line
				
				for (unsigned bk = 0; bk < booksCnt; bk++) {
					fin.getline (title, MAX_TITLE); 
					m_fullTitles.push_back (chomp(title));
				}

				for (unsigned bk = 0; bk < booksCnt; bk++) {
					fin.getline (title, MAX_TITLE); 
					m_shortTitles.push_back (chomp(title));
				}

				for (unsigned bk = 0; bk < booksCnt; bk++) {
					fin.getline (title, MAX_TITLE); 
					m_abbrTitles.push_back (chomp(title));
				}
				
				fin.close();
			}			
		}
		
		CBibleText::CBibleText (const char* compactTextPath, const char* compactPointPath) 
			throw (ios_base::failure)
		{
			m_compactPoint = new CFileVector<unsigned> (compactPointPath);
			
			m_textStream.open (compactTextPath, ios_base::binary);
			if (! m_textStream) 
				throw ios_base::failure (format("CBibleText::CBibleText(): error opening file %s", compactTextPath));  
		}
		
		string CBibleText::operator [] (int localVerseNum) 
			throw (out_of_range, ios_base::failure) 
		{
			unsigned filePos = (*m_compactPoint) [localVerseNum];
			m_textStream.seekg(filePos, ios_base::beg);
			
			unsigned char sz [2];
			m_textStream.read ((char*)sz, 2);
			if (m_textStream.gcount() < 2)
				throw ios_base::failure ("CBibleText::operator [](): I/O error");
				
			unsigned verseLen = sz[0] | (unsigned(sz[1]) << 8);
			
			char * verseText = new char [verseLen + 1];
			m_textStream.read (verseText, verseLen);
			if (m_textStream.gcount() < (int)verseLen)
				throw ios_base::failure ("CBibleText::operator [](): I/O error");

			string result (verseText, verseLen);
			delete [] verseText;
			
			return result;
		}
		
		CWordList::CWordList (const char* wordListPath, const char* wordPointPath, bool loadIntoMem) throw (ios_base::failure) 
			: m_loaded (false)
		{
			//TEMP
			if (! loadIntoMem)
				throw logic_error ("CWordList::CWordList(): loadIntoMem == false is not implemented yet");
			//END TEMP
		
			if (loadIntoMem) m_wordStream.open (wordListPath);
			else 			 m_wordStream.open (wordListPath, ios_base::binary);	
			
			if (! m_wordStream) 
				throw ios_base::failure (format("CWordList::CWordList(): error opening file %s", wordListPath));  
				
			if (! loadIntoMem) {
				m_wordPoint = new CFileVector<unsigned> (wordPointPath);				
				m_count = m_wordPoint->size();
			} else {
				char word [MAX_WORD_LEN + 1];
				while (m_wordStream.getline(word, MAX_WORD_LEN + 1, '\n')) 
					m_words.push_back (chomp(word));
					
				m_count = m_words.size();
				m_wordStream.close();
				
				m_wordPoint = NULL;
				m_loaded = true;
			}	
		}
		
		string CWordList::operator [] (unsigned wordIndex) throw (out_of_range, ios_base::failure) {
			if (wordIndex >= m_count)
				throw out_of_range ("CWordList::operator [](): wordIndex out of range");
				
			if (m_loaded)
				return m_words [wordIndex];

			char word [MAX_WORD_LEN + 3]; //+ CR, LF
			m_wordStream.seekg ((*m_wordPoint) [wordIndex], ios_base::beg);
			m_wordStream.read (word, MAX_WORD_LEN + 3);
			char* wordEnd = strstr (word, "\x0D\x0A");
			if (wordEnd == NULL)
				throw ios_base::failure ("CWordList::operator [](): I/O error");
				
			return string (word, wordEnd - word);
		}
		
		bool caseInsensitiveComp (const string& s1, const string& s2) {
			return rs::cyr_stricmp (s1.c_str(), s2.c_str()) < 0;		
		}

		int CWordList::indexOf (const string& word) const {
			if (m_loaded) {
				typedef vector<string>::const_iterator WordIterT;
				pair<WordIterT, WordIterT> foundI = 
					equal_range (m_words.begin(), m_words.end(), word, caseInsensitiveComp);
				
				//_log.printf ("wordIndex = %d\n", int(foundI.first - m_words.begin()));
				
				if (foundI.first == foundI.second) //not found
					return -1;
					
				return int(foundI.first - m_words.begin());
			}
			
			//TODO: Implement in case m_loaded == false; i.e. binary search in file
			return -1;
		}
		
		CWordList::~CWordList () {
			if (m_wordStream.is_open())
				m_wordStream.close();
			delete m_wordPoint;
		}
		
		void CConc::getVersesFor (unsigned wordIndex, vector<unsigned> & verseList) 
			throw (out_of_range)
		{
			if (wordIndex >= m_concPoint.size())
				throw out_of_range ("CConc::getVersesFor(): wordIndex out of range");
			
			verseList.clear();
			unsigned concIndex = m_concPoint [wordIndex];
			unsigned short count = m_conc [concIndex];
			while (count-- > 0) 
				verseList.push_back (m_conc [++concIndex] - 1); //-1 as VanyoG's format uses 1-based verse nums
		}
		
		CBibleInfo::CBibleInfo(const char* filePath, const char* defaults) {
			ifstream fin (filePath);
			char line [ITitles::MAX_TITLE];
			
			if (! fin)
				m_fullTitle = defaults;
			else {
				fin.getline (line, ITitles::MAX_TITLE, '\n');
				m_fullTitle = chomp(line);
			}

			if (! fin)
				m_shortTitle = m_fullTitle;
			else {
				fin.getline (line, ITitles::MAX_TITLE, '\n');
				m_shortTitle = chomp(line);
			}			
			
			if (! fin)
				m_abbreveatedTitle = m_shortTitle;
			else {
				fin.getline (line, ITitles::MAX_TITLE, '\n');
				m_abbreveatedTitle = chomp(line);
			}			
			
			if (fin) { //copyright defaults to ""
				while (fin.getline (line, ITitles::MAX_TITLE, '\n'))
					m_copyright += chomp(line) + '\n';
			}
			
			if (fin.is_open())
				fin.close();
		}
			
		CBible::CBible (const string& bibleFolder) throw (runtime_error)
		{
			try {
				string folder = bibleFolder;
				if (folder [folder.size() - 1] != '/')
					folder += "/";
					
				m_titles = new CTitles ((folder + "BibleTitles.txt").c_str());
				m_structure = new CBibleStructure ((folder + "BibleStructure.csv").c_str(),
												   m_titles->booksCount());
				m_diff = new CDiff ((folder + "_Diff_.txt").c_str());
				
				m_text = new CBibleText ((folder + "CompactText.bin").c_str(),
										(folder + "CompactPoint.bin").c_str());
				m_wordList = new CWordList ((folder + "WordList.txt").c_str(),
											(folder + "WordPoint.bin").c_str());
				m_conc = new CConc ((folder + "Conc.bin").c_str(),
									(folder + "ConcP.bin").c_str());

				m_info = new CBibleInfo ((folder + "_Info_.txt").c_str(), 
							static_cast<const char*>(QDir(folder).dirName()));
			} catch (exception & ex) {
				throw runtime_error (
					format ("CBible::CBible(): Error loading Bible (from folder %s);\nError message: \"%s\"",
							bibleFolder.c_str(), ex.what()));
			}			
		}
		
		CRef CBible::toLocal (const CRef & globalRef) const
			throw (out_of_range)
		{
			unsigned localBookID = m_titles->localID (globalRef.bookID());
			return m_diff->toLocal ( CRef (localBookID, 
											globalRef.chapter(), 
											globalRef.verse()));
		}
		
//		unsigned verseNum = m_structure->toNum (localRef);

		CRef CBible::toGlobal (const CRef & localRef) const
			throw (out_of_range)
		{
			unsigned globalBookID = m_titles->globalID(localRef.bookID());
			return m_diff->toGlobal ( CRef (globalBookID, 
											localRef.chapter(), 
											localRef.verse()));
		}
		
		string CBible::operator [] (const CRef & ref) const
			throw (out_of_range)
		{
			if (! ref.isGlobal() && ! ref.isLocal()) 
				throw out_of_range ("CBible::operator [](): ref type is not known");
				
			CRef localRef = ref;
			if (ref.isGlobal())
				localRef = this->toLocal(ref);
				
			unsigned verseNum = m_structure->toNum(localRef);
			return (*m_text) [verseNum];
		}
		
		void CBible::search (const string& searchText, bool wholeWord, bool caseSensitive,
								vector<CRef> & searchResult) const 
			throw (logic_error) 
		{
			if (wholeWord) {
				if (caseSensitive) //TODO: remove this restriction
					throw logic_error ("Whole words search may be only case insensitive");
			
				searchResult.clear();
				int wordIndex = m_wordList->indexOf (searchText);
				//_log.printf ("wordIndex = %d\n", wordIndex);
				if (wordIndex != -1) {
					vector<unsigned> verseNums;
					m_conc->getVersesFor (wordIndex, verseNums);
					
					//_log.printf ("SEARCH: \n");
					for (size_t v = 0; v < verseNums.size(); v++) {
						//_log.printf ("index %u, versenum %u, ", v, verseNums[v]);
						
						CRef localRef = m_structure->toRef (verseNums [v]);
						//_log.printf ("loclaref %s, ", localRef.toString(m_titles).c_str());
						
						searchResult.push_back (this->toGlobal(localRef));
						//_log.printf ("globalref %s\n", this->toGlobal(localRef).toString(this).c_str());
					}
				}
			} else {
				searchResult.clear();			
				if (caseSensitive) {
					for (unsigned v = 0; v < m_structure->totalVersesCount(); v++) {
						if ((* m_text)[v].find (searchText) != string::npos) {
							CRef localRef = m_structure->toRef (v);
							searchResult.push_back (this->toGlobal(localRef));
						}	
					}
				} else {
					for (unsigned v = 0; v < m_structure->totalVersesCount(); v++) {
						if (cyr_stristr((*m_text)[v].c_str(), searchText.c_str()) != NULL) {
							CRef localRef = m_structure->toRef (v);
							searchResult.push_back (this->toGlobal(localRef));
						}	
					}
				}
			}
		}
		
		string CBible::demoString() const
			throw (runtime_error) 
		{
			string demo;
			
			try {
				CRef ref = m_structure->toRef (31094);
				demo = format ("num(%u) -> ref(%u, %u, %u)",
					31094, ref.bookID(), ref.chapter(), ref.verse());
			} catch (exception & ex) {
				demo = string("Demo exception: ") + ex.what();
			}
				
			return demo;
		}
		
		CBible::~CBible () {
			delete m_conc; delete m_wordList; delete m_text;
			delete m_diff; delete m_structure; delete m_titles;
		}
		
		CBibleModules::CBibleModules (const string& modulesFolder) 
			throw (runtime_error) 
			: m_globalStructure(NULL)
		{
			m_modulesDir = modulesFolder;
			size_t lastIndex = m_modulesDir.size() - 1;
			if (m_modulesDir [lastIndex] == '/')
				m_modulesDir.erase (lastIndex, 1);
				
			QDir modulesDir (m_modulesDir);
			modulesDir.setFilter (QDir::Dirs);
			for (uint dirI = 0; dirI < modulesDir.count(); dirI++) {
				if (modulesDir[dirI].startsWith("."))
					continue;
			
				string biblePath = modulesDir.path () + "/" + modulesDir[dirI];
				m_bibleDirs.push_back (biblePath);
			}
		}
		
		const CBibleStructure* CBibleModules::globalStructure () const 
			throw (ios_base::failure)
		{
			if (m_globalStructure == NULL)
				m_globalStructure = new CBibleStructure (
					(m_modulesDir + "/BibleStructure.txt").c_str(), 
					77);
			return m_globalStructure;
		}
		
		const CBibleInfo* CBibleModules::info (unsigned bibleIndex) const 
			throw (out_of_range, ios_base::failure)
		{			
			if (bibleIndex >= m_bibleDirs.size())
				throw out_of_range ("CBibleModules::info(): index out of range");

			InfoMapT::iterator infoI = m_infos.find (bibleIndex);
			if (infoI != m_infos.end()) 
				return infoI->second;
			else {
				string bibleDir = m_bibleDirs[bibleIndex];
				CBibleInfo * info = new CBibleInfo ((bibleDir + "/_Info_.txt").c_str(), 
							static_cast<const char*>(QDir(bibleDir).dirName()));
				m_infos.insert (make_pair(bibleIndex, info));
				return info;
			}
		}		
		
		const CBible * CBibleModules::loadBible (unsigned bibleIndex) const 
			throw (out_of_range, ios_base::failure)
		{
			if (bibleIndex >= m_bibleDirs.size())
				throw out_of_range ("CBibleModules::loadBible(): index out of range");
			return new CBible (m_bibleDirs [bibleIndex]);
		}
		
		CBibleModules::~CBibleModules () {
			if (m_globalStructure != NULL)
				delete m_globalStructure;
				
			InfoMapT::iterator infoI = m_infos.begin();
			for (; infoI != m_infos.end(); infoI++)
				delete infoI->second;
		}
	}
}


