// Description: 
// Defines classes for accessing VanyoG Bible modules
//
// Author: Radoslav Atanasov <rsdevelop@gmail.com>, (C) 2005
// Copyright: See COPYING file that comes with this distribution

#ifndef VanyoG_Bible_H
#define VanyoG_Bible_H

#include "common.h"
#include "file_vector.h"
using namespace rs;

#include "vanyog_ref.h"

namespace BgBible {
	namespace VanyoG {
	
/*#if sizeof(unsigned) != 4
	#error "You need compiler with 4 byte unsigned"
#endif*/
	
		class CBibleStructure {
		protected:
			typedef vector<vector<unsigned> > VerseNumsT;
			VerseNumsT m_verses; //count of verses in each chapter; grouped by books
			
			unsigned m_totalVerseCount; 
			VerseNumsT m_startVerseNums; //num of first verse in each chapter
			vector<unsigned> m_bookStartVerseNums; //num of first verse in each book

			void CalcStartNums ();
						
			void ValidateBookID (unsigned bookID) const throw (out_of_range) {
				if (bookID >= m_verses.size())
					throw out_of_range ("CBibleSructure::ValidateBookID(): bookID out of range");
			}
			
			void ValidateChapter (unsigned bookID, unsigned chapter) const throw (out_of_range) {
				ValidateBookID (bookID);
				if (chapter >= m_verses[bookID].size())
					throw out_of_range ("CBibleSructure::ValidateChapterID(): chapter out of range");
			}
			
			void ValidateRef (const CRef& ref) const throw (out_of_range) {
				ValidateChapter (ref.bookID(), ref.chapter());
				if (ref.verse() >= m_verses [ref.bookID()] [ref.chapter()]) 
					throw out_of_range ("CBibleSructure::ValidateRef(): verse out of range");
			}
			
			void ValidateNum (unsigned num) const {
/*				unsigned booksCnt = m_verses.size();
				unsigned lastBookChapsCnt = m_verses [booksCnt - 1].size();
				unsigned lastChapVersesCnt = m_verses [booksCnt - 1][lastBookChapsCnt - 1];
				unsigned lastChapStartVerseNum = m_startVerseNums [booksCnt - 1] [lastBookChapsCnt - 1];*/
				
				if (num >= m_totalVerseCount /*lastChapStartVerseNum + lastChapVersesCnt*/)
					throw out_of_range ("CBibleSructure::ValidateNum(): verse number out of range");
			}
		public:
			CBibleStructure (const char* filePath, unsigned booksCount) throw (ios_base::failure);
			
			unsigned booksCount () const { 
				return m_verses.size(); 
			}
			
			unsigned chaptersCount (unsigned bookID) const throw (out_of_range) {
				ValidateBookID (bookID);
				return m_verses [bookID].size();
			}
			
			unsigned totalVersesCount () const { return m_totalVerseCount; }
			unsigned versesCount (unsigned bookID, unsigned chapter) const throw (out_of_range) 
			{
				ValidateChapter (bookID, chapter);
				return m_verses[bookID][chapter];
			}

			unsigned toNum (const CRef& ref) const throw (out_of_range);
			CRef toRef (unsigned num) const throw (out_of_range);
		};
		
// 		class CGlobalStructure : public CBibleSructure {
// 		public:
// 			explicit CGlobalStructure (const char* filePath, unsigned booksCount = 77) throw (runtime_error)
// 				: CBibleSructure (filePath, booksCount) 
// 			{
// 			}
// 		};
// 		
// 		class CLocalStructure : public CBibleSructure {
// 		public:
// 			explicit CLocalStructure (const char* filePath, unsigned booksCount) throw (runtime_error)
// 				: CBibleSructure (filePath, booksCount) 
// 			{
// 			}
// 		};
	
		class CDiff {
			typedef map<CRef, CRefDelta> DiffMapT;
			DiffMapT m_localToGlobal;
			DiffMapT m_globalToLocal;
			
			CRef convertWithMap (const CRef& ref, const DiffMapT& aMap);
		public:
			explicit CDiff (const char * filePath) throw (ios_base::failure);
			
			CRef toGlobal (const CRef& ref) { 
				return CRef (convertWithMap (ref, m_localToGlobal), true); 
			}
			
			CRef toLocal (const CRef& ref) { 
				return CRef (convertWithMap (ref, m_globalToLocal), false); 
			}
				
		};
		
		class CTitles : public ITitles {
			vector<unsigned> m_globalBookIDs;
			map<unsigned, unsigned> m_globalsToLocalsMap;
			vector<string> m_fullTitles, m_shortTitles, m_abbrTitles;
			
			void ValidateBookID (unsigned localBookID) const throw (out_of_range) {
				if (localBookID >= m_globalBookIDs.size())
					throw out_of_range ("CTitles::ValidateBookID(): bookID out of range");
			}
		public:
			explicit CTitles (const char* filePath) throw (ios_base::failure);
			
			unsigned booksCount () const {
				return m_globalBookIDs.size();
			}

			virtual string fullTitle (unsigned localBookID) const throw (out_of_range) {
				ValidateBookID (localBookID);
				return m_fullTitles [localBookID];
			}
			virtual string shortTitle (unsigned localBookID) const throw (out_of_range) {
				ValidateBookID (localBookID);
				return m_shortTitles [localBookID];
			}
			virtual string abbreviatedTitle (unsigned localBookID) const throw (out_of_range) {
				ValidateBookID (localBookID);
				return m_abbrTitles [localBookID];
			}
			
			bool hasBook (unsigned globalBookID) const {
				return m_globalsToLocalsMap.find (globalBookID) != m_globalsToLocalsMap.end();
			}
			
			unsigned localID (unsigned globalBookID) const throw (out_of_range) {
				map<unsigned, unsigned>::const_iterator mapping =
					m_globalsToLocalsMap.find (globalBookID);
				if (mapping == m_globalsToLocalsMap.end())
					throw out_of_range ("CTitles::localID() called for non existing book");
				return mapping->second;
			}
			
			unsigned globalID (unsigned localBookID) const throw (out_of_range) {
				ValidateBookID (localBookID);
				return m_globalBookIDs[localBookID];
			}
			
			virtual ~CTitles () {}
		};
		
		class CBibleText {
			ifstream m_textStream;
			CFileVector<unsigned> * m_compactPoint;
		public:
			CBibleText (const char* compactTextPath, const char* compactPointPath) throw(ios_base::failure);
			
			string operator [] (int localVerseNum) throw (out_of_range, ios_base::failure);
			
			~CBibleText () {
				m_textStream.close();
				delete m_compactPoint;
			}
		};
		
		class CWordList {
			unsigned m_count;
			bool m_loaded;
		
			CFileVector<unsigned> * m_wordPoint;
			ifstream m_wordStream;
				
			vector<string> m_words;
			
			enum {MAX_WORD_LEN = 40}; //I hope there isn't such a long word in the Bible :)
		public:
			explicit CWordList (const char* wordListPath, const char* wordPointPath, bool loadIntoMem = true) throw (ios_base::failure);
			
			unsigned count () const { return m_count; }
			string operator [] (unsigned wordIndex) throw (out_of_range, ios_base::failure);
			int indexOf (const string& word) const;
			
			~CWordList ();
		};
		
		class CConc {
			CFileVector<unsigned short> m_conc;
			CFileVector<unsigned> m_concPoint;
		public:
			CConc (const char* concPath, const char* concPointPath) throw (ios_base::failure)
				: m_conc(concPath), m_concPoint(concPointPath)
			{
			}
			
			void getVersesFor (unsigned wordIndex, vector<unsigned> & verseList) throw (out_of_range);
		};
		
		class CBibleInfo {
			string m_copyright;
			string m_fullTitle, m_shortTitle, m_abbreveatedTitle;
		public:
			explicit CBibleInfo (const char* filePath, const char* defaults = "");
			
			string copyright() const { return m_copyright; }
			
			string fullTitle() const { return m_fullTitle; }
			string shortTitle() const { return m_shortTitle; }
			string abbreveatedTitle() const { return m_abbreveatedTitle; }
		};
		
		class CBible : public ITitles {
		protected:
			CBibleStructure * m_structure;
			CDiff * m_diff;
			CTitles * m_titles;
			
			CBibleText * m_text;
			CWordList * m_wordList;
			CConc * m_conc;
			
			CBibleInfo * m_info;
		protected:
			//methods	
		public:
			explicit CBible (const string& bibleFolder) throw (runtime_error);

			//public components			
			const CBibleInfo * info() const { return m_info; }
			const CTitles * titles () const { return m_titles; }
			
			//structure
			unsigned booksCount () const {	
				return m_structure->booksCount (); 
			}
			unsigned chaptersCount (unsigned bookID) const throw (out_of_range) {
				return m_structure->chaptersCount(bookID);
			}
			unsigned versesCount (unsigned bookID, unsigned chapter) const throw (out_of_range) {
				return m_structure->versesCount(bookID, chapter);
			}
			
			//global<->local conversion
			CRef toLocal (const CRef& globalRef) const throw (out_of_range);
			CRef toGlobal (const CRef& localRef) const throw (out_of_range);
			
			//titles
			bool hasBook (unsigned globalBookID) const {
				return m_titles->hasBook(globalBookID);
			}
			virtual string fullTitle (unsigned globalBookID) const throw (out_of_range) {
				if (! m_titles->hasBook(globalBookID)) 
					return "";
				return m_titles->fullTitle (m_titles->localID (globalBookID));
			}
			virtual string shortTitle (unsigned globalBookID) const throw (out_of_range) {
				if (! m_titles->hasBook(globalBookID)) 
					return "";
				return m_titles->shortTitle (m_titles->localID (globalBookID));
			}
			virtual string abbreviatedTitle (unsigned globalBookID) const throw (out_of_range) {
				if (! m_titles->hasBook(globalBookID)) 
					return "";
				return m_titles->abbreviatedTitle (m_titles->localID (globalBookID));
			}
		
			//getting text
			string operator [] (const CRef& ref) const throw (out_of_range);
			
			//word list
			unsigned wordsCount () const { 
				return m_wordList->count(); 
			}
			string getWord (unsigned wordIndex) const throw (out_of_range, ios_base::failure) {
				return (*m_wordList) [wordIndex];
			}
			int indexOfWord (const string& word) const {
				return m_wordList->indexOf (word);
			}
			
			//search
			void search (const string& text, bool wholeWord, bool caseSensitive,
				vector<CRef> & searchResult) const throw (logic_error);

			//testing method
			string demoString () const throw (runtime_error); 
						
			virtual ~CBible ();
		};
		
		class CBibleModules {
			string m_modulesDir;
			vector<string> m_bibleDirs;
			
			typedef map<unsigned, CBibleInfo*> InfoMapT;
			mutable InfoMapT m_infos;
			
			mutable CBibleStructure * m_globalStructure;
		public:
			explicit CBibleModules (const string& modulesFolder) throw (runtime_error);
			
			unsigned count () const { return m_bibleDirs.size(); }
			
			const CBibleStructure* globalStructure () const throw (ios_base::failure);
			const CBibleInfo* info (unsigned bibleIndex) const throw (out_of_range, ios_base::failure);
			const CBible * loadBible (unsigned bibleIndex) const throw (out_of_range, ios_base::failure);
			
			virtual ~CBibleModules ();
		};
	}
}

#endif
